# Copyright (C) 2023 Adam Lugowski.
# Use of this source code is governed by the BSD 2-clause license found in the LICENSE.txt file.
# SPDX-License-Identifier: BSD-2-Clause

"""
Removed in favor of Jupyter extension:

%load_ext matrepr.html
"""

raise DeprecationWarning("Please switch to Jupyter extension: %load_ext matrepr.html")
